use deimos::vm::dm_type::DmType;
use deimos::vm::lib::{DmConstant, DmLib};
use deimos::vm::object_type::{DmField, DmFn, DmImplementation, DmInterface, DmMethod};
use deimos::vm::op_codes::{
    add_alloc, add_alloc_raw_from, add_invoke_fn, add_mov_const, add_mov_register_to,
    add_mov_size_of, add_multiply, add_platform_call,
};
use deimos::vm::{call_fn, DvmContext, DvmState};
use std::rc::Rc;

fn main() {
    // Goal:
    // - write a single lib with a main()
    // - call the main fn
    // fn main() { println "Hello, World!" }

    // std/unsafe/mem lib
    let mut mem_lib = DmLib::new("std/unsafe/mem");

    // std::unsafe::Pointer
    let pointer_int = DmInterface::new("std::unsafe::Pointer", "Pointer");
    let pointer_int_rc = Rc::new(pointer_int);

    // std::unsafe::PointerImpl : Pointer
    let mut pointer_impl = DmImplementation::new(
        "std::unsafe:PointerImpl",
        "PointerImpl",
        Some(pointer_int_rc.clone()),
    );

    let pointer_impl_raw_address = DmField::new("raw_address", DmType::Long, 0);
    let pointer_impl_size = DmField::new("size", DmType::Long, 8);

    // std::unsafe::PointerImpl::_ctor_0(
    //   r0: self
    //   r1: size Long
    // )
    // r2: raw_address Long
    let mut pointer_impl_ctor_0_code: Vec<u8> = Vec::new();
    add_alloc_raw_from(&mut pointer_impl_ctor_0_code, 2, 1);
    add_mov_register_to(
        &mut pointer_impl_ctor_0_code,
        0,
        pointer_impl_raw_address.data_offset() as u32,
        2,
    );
    add_mov_register_to(
        &mut pointer_impl_ctor_0_code,
        0,
        pointer_impl_size.data_offset() as u32,
        1,
    );

    let pointer_impl_ctor_0_fn = DmFn::new(
        "std::unsafe::PointerImpl::_ctor_0",
        "_ctor_0",
        pointer_impl_ctor_0_code,
        3,
        None,
    );

    let pointer_impl_ctor_0_method = DmMethod::new(pointer_impl_ctor_0_fn, None);

    pointer_impl.fields.push(pointer_impl_raw_address);
    pointer_impl.fields.push(pointer_impl_size);
    pointer_impl
        .methods
        .push(Rc::new(pointer_impl_ctor_0_method));

    let unsafe_pointer_impl_rc = Rc::new(pointer_impl);

    mem_lib.interfaces.push(pointer_int_rc.clone());
    mem_lib.implementations.push(unsafe_pointer_impl_rc.clone());

    // std::unsafe::alloc(
    //   r0: size Long
    // )
    // r0: size Long
    // r1: Pointer object
    // r2: Void from PointerImpl::_ctor_0
    // @return r1
    let mut alloc_fn_code: Vec<u8> = Vec::new();
    add_alloc(
        &mut alloc_fn_code,
        1,
        unsafe_pointer_impl_rc.size_in_bytes() as u32,
        "std::unsafe::PointerImpl",
    );
    add_invoke_fn(
        &mut alloc_fn_code,
        "std::unsafe::PointerImpl::_ctor_0",
        2,
        &[1u8, 0u8],
    );

    let alloc_fn = DmFn::new(
        "std::unsafe::mem::alloc",
        "alloc",
        alloc_fn_code,
        3,
        Some(1),
    );

    mem_lib.functions.push(Rc::new(alloc_fn));

    // std/core/array lib
    let mut array_lib = DmLib::new("std/core/array");

    // std::core::Array
    let array_int = DmInterface::new("std::core::Array", "Array");
    let array_int_rc = Rc::new(array_int);

    // std::core::ArrayImpl
    let mut array_impl = DmImplementation::new(
        "std::core::ArrayImpl",
        "ArrayImpl",
        Some(array_int_rc.clone()),
    );

    let array_impl_pointer_fld = DmField::new("pointer", DmType::Pointer, 0);
    let array_impl_length_fld = DmField::new("length", DmType::Int, 8);

    // std::core::Array::_ctor_0(
    //   r0: self
    //   r1: TypeRef element_type
    //   r2: Int length
    // )
    // r3: r1::size()
    // r4: r3 * r1
    // r5: Pointer allocated pointer
    let mut array_impl_ctor_0_bytecode: Vec<u8> = Vec::new();

    add_mov_size_of(&mut array_impl_ctor_0_bytecode, 3, 1);
    add_multiply(&mut array_impl_ctor_0_bytecode, 4, 3, 1);
    add_invoke_fn(
        &mut array_impl_ctor_0_bytecode,
        "std::unsafe::mem::alloc",
        5,
        &[4],
    );
    add_mov_register_to(
        &mut array_impl_ctor_0_bytecode,
        0,
        array_impl_pointer_fld.data_offset() as u32,
        5,
    );
    add_mov_register_to(
        &mut array_impl_ctor_0_bytecode,
        0,
        array_impl_length_fld.data_offset() as u32,
        2,
    );

    let array_impl_ctor_0_fn = DmFn::new(
        "std::core::ArrayImpl::_ctor_0",
        "_ctor_0",
        array_impl_ctor_0_bytecode,
        6,
        None,
    );
    let array_impl_ctor_0_method = DmMethod::new(array_impl_ctor_0_fn, None);
    array_impl.methods.push(Rc::new(array_impl_ctor_0_method));

    // Add Array and ArrayImpl to array lib
    array_lib.interfaces.push(array_int_rc.clone());
    array_lib.implementations.push(Rc::new(array_impl));

    // std::core::String
    let mut string_lib = DmLib::new("std/core/string");

    let string_int = DmInterface::new("std::core::String", "String");

    let mut string_impl = DmImplementation::new("std::core::StringImpl", "StringImpl", None);

    let bytes_field = DmField::new("bytes", DmType::Pointer, 0);

    // std::core::String::_ctor_0(
    //   r0: self
    //   r1: DvmPointer to Array<Byte>
    // )
    let mut string_ctor_0_bytecode: Vec<u8> = Vec::new();
    add_mov_register_to(
        &mut string_ctor_0_bytecode,
        0,
        bytes_field.data_offset() as u32,
        1,
    );

    let string_ctor_0_fn = DmFn::new(
        "std::core::StringImpl::_ctor_0",
        "_ctor_0",
        string_ctor_0_bytecode,
        2,
        None,
    );

    let string_ctor_0_method = DmMethod::new(string_ctor_0_fn, None);

    string_impl.fields.push(bytes_field);
    string_impl.methods.push(Rc::new(string_ctor_0_method));
    let core_string_impl_size = string_impl.size_in_bytes();

    string_lib.interfaces.push(Rc::new(string_int));
    string_lib.implementations.push(Rc::new(string_impl));

    // greeting lib
    let mut greeting_lib = DmLib::new("greeting");
    greeting_lib
        .constants
        .push(DmConstant::String("Hello, World!".to_string()));

    let mut main_byte_code = Vec::new();

    // 1. Move constant: r0 receives DvmValue::Pointer to ArrayImpl<Byte>
    // 2. Allocate for std::core::StringImpl into r1
    // 3. Call StringImpl::_ctor_0(r0) -> r2
    // 4. Platform call std::core::println(r1) -> r3
    add_alloc(
        &mut main_byte_code,
        0,
        core_string_impl_size as u32,
        "std::core::StringImpl",
    );
    add_mov_const(&mut main_byte_code, 1, "greeting", 0);
    add_invoke_fn(
        &mut main_byte_code,
        "std::core::StringImpl::_ctor_0",
        2,
        &[0, 1],
    );
    add_platform_call(&mut main_byte_code, "std::core::println", 3, 1, &vec![0u8]);

    let main_dm_fn = DmFn::new("main", "main", main_byte_code, 2, None);

    greeting_lib.functions.push(Rc::new(main_dm_fn));

    let mut state = DvmState::new();
    let mut context = DvmContext::new();
    context.load(vec![
        Rc::new(greeting_lib), 
        Rc::new(string_lib), 
        Rc::new(array_lib), 
        Rc::new(mem_lib)
    ]);

    let main_fn = context.fn_by_fqn("main").unwrap();
    call_fn(&mut state, &context, &main_fn, vec![]);
}
