pub mod dm_type;
pub mod dvm_value;
pub mod lib;
pub mod mem;
pub mod object_type;
pub mod op_codes;
pub mod platform;
pub mod util;

use crate::vm::dvm_value::DvmValue;
use crate::vm::lib::{DmConstant, DmLib};
use crate::vm::mem::{get_field_value, DmAllocObject};
use crate::vm::object_type::DmFn;
use op_codes::*;
use std::alloc::{alloc, dealloc, Layout};
use std::collections::HashMap;
use std::rc::Rc;

pub type PlatformFunction =
    fn(args: Vec<DvmValue>, state: &mut DvmState, context: &DvmContext) -> DvmValue;

enum CallFrame {
    PlatformCall(CallFrameInfo),
    DeimosCall(CallFrameInfo),
}

struct CallFrameInfo {
    fn_fqn: String,
}

pub struct DvmContext {
    libs: Vec<Rc<DmLib>>,
    functions: HashMap<String, Rc<DmFn>>,
    platform_functions: HashMap<String, Rc<PlatformFunction>>,
}

impl DvmContext {
    pub fn new() -> Self {
        DvmContext {
            libs: Vec::new(),
            functions: HashMap::new(),
            platform_functions: HashMap::new(),
        }
    }

    pub fn load(&mut self, libs: Vec<Rc<DmLib>>) {
        for lib in libs {
            self.libs.push(lib.clone());
            for lib_fn in &lib.functions {
                self.functions
                    .insert(lib_fn.fqn().to_string(), lib_fn.clone());
            }
            for interface_function in lib
                .interfaces
                .iter()
                .flat_map(|interface| interface.get_functions())
            {
                self.functions.insert(
                    interface_function.fqn().to_string(),
                    interface_function.clone(),
                );
            }
            for implementation_function in lib
                .implementations
                .iter()
                .flat_map(|implementation| &implementation.functions)
            {
                self.functions.insert(
                    implementation_function.fqn().to_string(),
                    implementation_function.clone(),
                );
            }
            for method in lib
                .implementations
                .iter()
                .flat_map(|implementation| &implementation.methods)
            {
                self.functions.insert(
                    method.dm_fn().fqn().to_string(),
                    method.dm_fn().clone(),
                );
            }
        }
    }

    pub fn fn_by_fqn(&self, fqn: &str) -> Option<Rc<DmFn>> {
        self.functions.get(fqn).cloned()
    }

    pub fn platform_fn_by_fqn(&self, fqn: &str) -> Option<Rc<PlatformFunction>> {
        self.platform_functions.get(fqn).cloned()
    }
}

pub struct DvmState {
    call_stack: Vec<CallFrame>,
    registers: Vec<DvmValue>,
    register_state_stack: Vec<Vec<DvmValue>>,
}

impl DvmState {
    pub fn new() -> Self {
        DvmState {
            call_stack: Vec::new(),
            registers: Vec::new(),
            register_state_stack: Vec::new(),
        }
    }
}

pub fn call_fn(
    state: &mut DvmState,
    context: &DvmContext,
    dm_fn: &DmFn,
    args: Vec<DvmValue>,
) -> DvmValue {
    // save current state
    state.call_stack.push(CallFrame::DeimosCall(CallFrameInfo {
        fn_fqn: dm_fn.fqn().to_string(),
    }));
    state.register_state_stack.push(state.registers.clone());

    // zero registers and make sure there are enough for dm_function
    state.registers.clear();
    if state.registers.len() < dm_fn.number_used_registers() {
        state
            .registers
            .resize(dm_fn.number_used_registers(), DvmValue::Uninit);
    }

    // push args
    for arg in args {
        state.registers.push(arg);
    }

    // run the byte code
    run_byte_code(state, context, &dm_fn.byte_code());

    // restore state
    state.registers = state.register_state_stack.pop().unwrap();
    state.call_stack.pop();

    // get return value
    if let Some(return_register) = dm_fn.return_register() {
        state
            .registers
            .get(return_register as usize)
            .map(|m| m.clone())
            .unwrap()
    } else {
        DvmValue::Void
    }
}

macro_rules! next_8 {
    ( $byte_code: expr, $T: ident ) => {
        $byte_code.next().unwrap() as $T
    };
}

macro_rules! next_32_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
    };
}

macro_rules! next_64_le {
    ( $byte_code: expr, $T: ident ) => {
        next_8!($byte_code, $T)
            + (next_8!($byte_code, $T) << 8)
            + (next_8!($byte_code, $T) << 16)
            + (next_8!($byte_code, $T) << 24)
            + (next_8!($byte_code, $T) << 32)
            + (next_8!($byte_code, $T) << 40)
            + (next_8!($byte_code, $T) << 48)
            + (next_8!($byte_code, $T) << 56)
    };
}

macro_rules! read_array_u8 {
    ( $iter: expr ) => {{
        let arr_length = next_8!($iter, u8);
        let mut arr = Vec::with_capacity(arr_length as usize);
        let mut i = 0;
        while i < arr_length {
            arr.push(next_8!($iter, u8));
            i += 1;
        }
        arr
    }};
}

macro_rules! read_string {
    ( $iter: expr ) => {{
        let str_length = next_32_le!($iter, usize);
        let mut str_raw: Vec<u8> = Vec::new();
        let mut i = 0;
        while i < str_length {
            str_raw.push($iter.next().unwrap());
            i += 1;
        }
        String::from_utf8(str_raw).unwrap()
    }};
}

macro_rules! write_string {
    ( $raw_ptr: expr, $s: expr ) => {
        for (i, b) in $s.bytes().enumerate() {
            unsafe {
                $raw_ptr.add(i).write(b);
            }
        }
    };
}

macro_rules! write_bytes {
    ( $ptr: expr, $offset: expr, $src: expr ) => {
        for (i, &b) in $src.iter().enumerate() {
            unsafe {
                $ptr.add($offset + i).write(b);
            }
        }
    };
}

pub fn run_byte_code(state: &mut DvmState, context: &DvmContext, byte_code: &[u8]) {
    let mut iter = byte_code.iter().cloned();
    while let Some(op_code) = iter.next() {
        match op_code {
            MOV_INT => {
                let target_register = next_8!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;
                state
                    .registers
                    .insert(target_register, DvmValue::Int(operand));
            }
            MOV_LONG => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as i64;
                state
                    .registers
                    .insert(target_register, DvmValue::Long(operand));
            }
            MOV_DOUBLE => {
                let target_register = next_8!(iter, usize);
                let operand = next_64_le!(iter, u64) as f64;
                state
                    .registers
                    .insert(target_register, DvmValue::Double(operand));
            }
            MOV_REGISTER => {
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let source_value = state.registers.get(source_register).unwrap();
                state
                    .registers
                    .insert(target_register, source_value.clone())
            }
            ALLOC => {
                let target_register = next_8!(iter, usize);
                let alloc_size = next_32_le!(iter, usize);
                let impl_name = read_string!(iter);

                let implementation = context
                    .libs
                    .iter()
                    .find_map(|lib| {
                        lib.implementations
                            .iter()
                            .find(|implementation| implementation.fqn == impl_name)
                    })
                    .expect(&format!("Implementation not found: {}", impl_name));

                let dm_alloc_object = DmAllocObject::new(alloc_size, implementation.clone());
                let dvm_pointer = DvmValue::Pointer(Rc::new(dm_alloc_object));
                state.registers.insert(target_register, dvm_pointer);
            }
            DEALLOC => {
                let target_register = next_8!(iter, usize);
                let dm_alloc_object = state.registers.remove(target_register).expect_pointer();
                drop(dm_alloc_object); // explicit
            }
            MOV_INT_TO => {
                let target_register = next_8!(iter, usize);
                let offset = next_32_le!(iter, usize);
                let operand = next_32_le!(iter, u32) as i32;

                let dm_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_pointer();
                write_bytes!(dm_alloc_object.data, offset, operand.to_ne_bytes());
            }
            MOV_LONG_TO => {
                todo!()
            }
            MOV_DOUBLE_TO => {
                todo!()
            }
            MOV_REGISTER_TO => {
                // Moves the value from the source register to the object pointed to by the target
                // register, at the given offset.
                let target_register = next_8!(iter, usize);
                let offset = next_32_le!(iter, usize);
                let source_register = next_8!(iter, usize);

                let target_alloc_object = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_pointer();

                let source_value = state.registers.get(source_register).unwrap();
                match source_value {
                    DvmValue::Int(i) => {
                        write_bytes!(target_alloc_object.data, offset, i.to_ne_bytes());
                    }
                    DvmValue::Byte(b) => unsafe {
                        target_alloc_object.data.add(offset).write(*b);
                    },
                    DvmValue::Long(l) => {
                        write_bytes!(target_alloc_object.data, offset, l.to_ne_bytes());
                    }
                    DvmValue::Double(d) => {
                        write_bytes!(target_alloc_object.data, offset, d.to_ne_bytes());
                    }
                    DvmValue::Boolean(b) => unsafe {
                        target_alloc_object.data.add(offset).write(*b as u8);
                    },
                    DvmValue::Pointer(source_alloc_object) => {
                        let source_ptr = Rc::into_raw(source_alloc_object.clone());
                        write_bytes!(
                            target_alloc_object.data,
                            offset,
                            (source_ptr as usize).to_ne_bytes()
                        );
                    }
                    DvmValue::Uninit => {
                        panic!("Cannot move DvmValue::Uninit to object.")
                    }
                    DvmValue::Void => {
                        panic!("Cannot move DvmValue::Void to object.")
                    }
                }
            }
            MOV_CONST => {
                // TODO: reduce the complexity of this.
                // Perhaps the target register should just contain a ptr where the data is written.

                // Decode
                let target_register = next_8!(iter, usize); // must contain an allocated StringImpl
                let lib_name = read_string!(iter);
                let const_id = next_8!(iter, usize);
                let pointer_impl_register = next_8!(iter, usize);
                let byte_array_impl_register = next_8!(iter, usize);

                // Get constant
                let Some(lib) = context.libs.iter().find(|lib| lib.name == lib_name) else {
                    panic!("Could not find lib with name: {}", lib_name);
                };
                let constant = lib.constants.get(const_id).unwrap();
                let DmConstant::String(s) = constant;

                // Get allocated PointerImpl
                let pointer_impl = state
                    .registers
                    .get(pointer_impl_register)
                    .unwrap()
                    .expect_pointer();

                // Init PointerImpl object
                let pointer_impl_ctor_0_fn = context
                    .functions
                    .get("std::unsafe::PointerImpl::_ctor_0")
                    .expect("Could not find std::unsafe::PointerImpl::_ctor_0");
                call_fn(
                    state,
                    context,
                    pointer_impl_ctor_0_fn,
                    vec![
                        DvmValue::Pointer(pointer_impl.clone()),
                        DvmValue::Long(s.len() as i64),
                    ],
                );

                // Get std::unsafe::PointerImpl.raw_address field as *mut u8
                let raw_ptr = unsafe {
                    get_field_value(
                        pointer_impl
                            .implementation
                            .get_field("raw_address", &pointer_impl)
                            .expect("Could not get PointerImpl.raw_address field."),
                        &pointer_impl,
                    )
                }
                .expect_long() as usize as *mut u8;

                // Write the constant bytes to the raw_ptr
                write_string!(raw_ptr, s);

                // Get allocated ArrayImpl<Byte> from byte_array_impl_register
                let byte_array_impl = state
                    .registers
                    .get(byte_array_impl_register)
                    .unwrap()
                    .expect_pointer();

                // Init ArrayImpl<Byte>
                let array_ctor_0_fn = context
                    .functions
                    .get("std::core::ArrayImpl::_ctor_1")
                    .expect("Could not find std::core::ArrayImpl::_ctor_1");
                call_fn(
                    state,
                    context,
                    array_ctor_0_fn,
                    vec![
                        DvmValue::Pointer(byte_array_impl.clone()), // self
                        DvmValue::Pointer(pointer_impl.clone()),    // PointerImpl
                        DvmValue::Int(s.len() as i32),              // length
                    ],
                );

                // Get allocated StringImpl from string_impl_register
                let string_impl = state
                    .registers
                    .get(target_register)
                    .unwrap()
                    .expect_pointer();

                // Init StringImpl
                let string_impl_ctor_0_fn = context
                    .functions
                    .get("std::core::StringImpl::_ctor_0")
                    .expect("Could not find std::core::StringImpl::_ctor_0");
                call_fn(
                    state,
                    context,
                    string_impl_ctor_0_fn,
                    vec![
                        DvmValue::Pointer(string_impl.clone()),
                        DvmValue::Pointer(byte_array_impl.clone()),
                    ],
                );
            }
            ALLOC_RAW => {
                // Allocates a raw number of bytes, with the number of bytes determined by the value
                // of the source register. The address is stored in the target register as a
                // DvmValue::Long.
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let number_of_bytes =
                    state.registers.get(source_register).unwrap().expect_long() as usize;
                let layout = Layout::from_size_align(number_of_bytes, 0).unwrap();
                let ptr = unsafe { alloc(layout) };
                state
                    .registers
                    .insert(target_register, DvmValue::Long(ptr as i64))
            }
            DEALLOC_RAW => {
                // Deallocates a raw number of bytes at an address, with the number of bytes
                // determined by the value in the source register, and the address stored in the
                // target register.
                let target_register = next_8!(iter, usize);
                let source_register = next_8!(iter, usize);
                let number_of_bytes =
                    state.registers.get(source_register).unwrap().expect_long() as usize;

                let layout = Layout::from_size_align(number_of_bytes, 0).unwrap();
                let ptr =
                    state.registers.get(target_register).unwrap().expect_long() as usize as *mut u8;
                unsafe {
                    dealloc(ptr, layout);
                }
                state.registers.insert(target_register, DvmValue::Uninit);
            }
            PLATFORM_CALL => {
                // Calls a platform function. The result of the platform call is stored in the
                // specified return register.
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let platform_function = context
                    .platform_functions
                    .get(&symbol_name)
                    .unwrap()
                    .clone();
                state
                    .call_stack
                    .push(CallFrame::PlatformCall(CallFrameInfo {
                        fn_fqn: symbol_name,
                    }));
                let call_result = platform_function(args, state, context);
                state.registers.insert(return_register, call_result);
                state.call_stack.pop();
            }
            INVOKE_FN => {
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let dm_fn = context.functions.get(&symbol_name).unwrap();
                let call_result = call_fn(state, context, dm_fn, args);
                state.registers.insert(return_register, call_result);
            }
            INVOKE_VIRTUAL => {
                let symbol_name = read_string!(iter);
                let return_register = next_8!(iter, usize);
                let arg_registers = read_array_u8!(iter);

                let mut args = Vec::with_capacity(arg_registers.len());
                for arg_register in arg_registers {
                    let value = state.registers.get(arg_register as usize).unwrap();
                    args.push(value.clone());
                }

                let self_obj = args.get(0).unwrap().expect_pointer();
                let method = self_obj
                    .implementation
                    .get_method(&symbol_name, &self_obj)
                    .expect(&format!("Could not find method: {}", symbol_name));
                let dm_fn = method.dm_fn();

                let call_result = call_fn(state, context, &dm_fn, args);
                state.registers.insert(return_register, call_result);
            }
            INVOKE_DYNAMIC => {
                unimplemented!("INVOKE_DYNAMIC is not yet supported and may never be.")
            }
            MOV_SIZE_OF => {
                todo!()
            }
            MULTIPLY => {
                todo!()
            }
            op_code => {
                panic!("Invalid op_code: {}", op_code);
            }
        }
    }
}

#[cfg(test)]
mod run_code_tests {
    use super::*;

    macro_rules! assert_register {
        ( $expected: expr, $state: expr, $register_number: expr ) => {
            assert_eq!(
                $expected,
                $state.registers.get($register_number).unwrap().clone()
            );
        };
    }

    fn setup() -> (DvmState, DvmContext) {
        (DvmState::new(), DvmContext::new())
    }

    #[test]
    fn mov_1_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 1);
        let (mut state, context) = setup();
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(1), state, 0);
    }

    #[test]
    fn move_65535_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0xffff);
        let (mut state, context) = setup();
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(0xffff), state, 0);
    }

    #[test]
    fn move_int_max_as_int() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 0, 0x0fff_ffff);
        let (mut state, context) = setup();
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(0x0fff_ffff), state, 0);
    }

    #[test]
    fn move_register() {
        let mut code = Vec::new();
        add_mov_int(&mut code, 1, 1);
        add_mov_register(&mut code, 0, 1);
        let (mut state, context) = setup();
        state.registers.resize(2, DvmValue::Uninit);
        run_byte_code(&mut state, &context, &code);
        assert_register!(DvmValue::Int(1), state, 0);
    }
}
